// ==========================================================
//
// CBP_Hexapod_1.c
//
// Hexapod robot following Campos paper
//
// CBP 01-12-20
//
// ==========================================================

#include <stdio.h>
#include <webots/motor.h>
#include <webots/robot.h>
 
// ODE Solvers
#include "rkf45.h"
#include "CBP_VderPol.h"
#include "CBP_Helper_H.h"

#define TIME_STEP 16
#define NUM_MOTORS 12

#define NEQN 12

//#define NUM_STATES 6

//#define FRONT +0.7
//#define BACK -0.7
#define UP -0.015
#define DOWN 0.015

#define DEBUG false

// defines for the ODEs
#define LFx 0
#define LFz 1
#define RFx 2
#define RFz 3
#define LMx 4
#define LMz 5
#define RMx 6
#define RMz 7
#define LRx 8
#define LRz 9
#define RRx 10
#define RRz 11

int main() {

  // rkf stuff
  double abserr;
  int flag;
  double relerr;
  double t;
  double t_out;
  double y[NEQN];
  double dydt[NEQN];
  
  double pi = M_PI;
  
  // params
  double alpha = 5;
  double beta;
  double mu = 1;
  double omegaSW = 2*pi;
  double a = 2.0;
  double theta_12, theta_13, theta_24, theta_34, theta_35, theta_46, theta_56;
   
  double phaseLF, phaseRF, phaseLM, phaseRM, phaseLR, phaseRR;
  double phaseLF_prev, phaseRF_prev, phaseLM_prev, phaseRM_prev, phaseLR_prev, phaseRR_prev;
  
  FILE* logFile;
  
  const char *MOTOR_NAMES[NUM_MOTORS] = {"hip_motor_r0",  "hip_motor_r1",  "hip_motor_r2",  "hip_motor_l0",
                                         "hip_motor_l1",  "hip_motor_l2",  "knee_motor_r0", "knee_motor_r1",
                                         "knee_motor_r2", "knee_motor_l0", "knee_motor_l1", "knee_motor_l2"};
  WbDeviceTag motors[NUM_MOTORS];

  int i;

  wb_robot_init();
    
  logFile = fopen("Robot.m","w");
  writeOctaveHead(logFile);

  for (i = 0; i < NUM_MOTORS; i++) {
    motors[i] = wb_robot_get_device(MOTOR_NAMES[i]);
    if (!motors[i])
      printf("could not find motor: %s\n", MOTOR_NAMES[i]);
  }

 // Calculate the epsilon
  abserr = 100*sqrt ( r8_epsilon ( ) );
  relerr = 100*sqrt ( r8_epsilon ( ) );
  printf("Root epsilon: %e\n",abserr);
  
  flag = 1;

  // Set some initial conditions
  for(i=0; i < 12; i++)
    y[i] = 0.1;
  
  // Ripple
  theta_12 = -pi;
  theta_13 = -3*pi/2.0;
  theta_24 =  pi/2.0;
  theta_34 =  pi;
  theta_35 =  pi/2.0;
  theta_46 =  pi/2.0;
  theta_56 =  pi;
  beta = 3.0/4.0;  
  
    // Metachronal
  theta_12 = -pi;
  theta_13 = -pi/3;
  theta_24 = -pi/3;
  theta_34 = -pi;
  theta_35 = -pi/3;
  theta_46 = -pi/3;
  theta_56 = -pi;
  beta = 5.0/6.0;   
  
  // tripod
theta_12 = -pi;
theta_13 = -pi;
theta_24 =  pi;
theta_34 =  pi;
theta_35 =  pi;
theta_46 = -pi;
theta_56 = -pi;
beta = 3.0/6.0;  
   
  
  setParams(alpha, beta, mu, omegaSW, a);
  setThetas(theta_12, theta_13, theta_24, theta_34, theta_35, theta_46, theta_56);  
  
  // Initialize the ODE Solver                 
  t = 0.0;
  t_out = TIME_STEP/1000.0; 
  vdPol3 ( t, y,  dydt );
    
  while (wb_robot_step(TIME_STEP) != -1) {
                      
    flag = r8_rkf45 ( vdPol3, NEQN, y, dydt, &t, t_out, &relerr, abserr, flag );
      
      
    if(DEBUG) printf ( "%4d  %12f  %12f  %12f  %12f  %12f %12f %12f\n",
     flag, t, y[1], y[3], y[5], y[7], y[9],y[11] );
 
    t_out = t + TIME_STEP/1000.0;
    
    if(flag == 4) {
      printf("WARNING: flag 4 at t = %f\n",t);
      flag = 2;
    }
    if(flag == 7) {
      printf("WARNING: flag 7 at t = %f\n",t);
      flag = 2; 
    }   
    if(flag == 6) {
      abserr *=1.1;
      relerr *=1.1;
      flag = 2;
      printf("WARNING: flag 6: abserr = %f\n",abserr);
    }
    
    phaseLF_prev = phaseLF;
    phaseRF_prev = phaseRF;
    phaseLM_prev = phaseLM;
    phaseRM_prev = phaseRM;
    phaseLR_prev = phaseLR;
    phaseRR_prev = phaseRR;

    phaseLF = y[LFx];
    phaseRF = y[RFx];
    phaseLM = y[LMx];
    phaseRM = y[RMx];	
    phaseLR = y[LRx];
    phaseRR = y[RRx];      

    writeData6ToOctave(logFile,t,phaseLF,phaseRF,phaseLM,phaseRM,phaseLR,phaseRR);

    // Code to identify motors and joints
    //for(i=0;i<6;i++)
    //  wb_motor_set_position(motors[i], 0.0);
    //  wb_motor_set_position(motors[10], 0.01);
    double scaling = 0.2;

    wb_motor_set_position(motors[2], scaling*y[LFx]);
    wb_motor_set_position(motors[5], scaling*y[RFx]);
    wb_motor_set_position(motors[1], scaling*y[LMx]);
    wb_motor_set_position(motors[4], scaling*y[RMx]);
    wb_motor_set_position(motors[0], scaling*y[LRx]);
    wb_motor_set_position(motors[3], scaling*y[RRx]);
    
    double velyLF = phaseLF - phaseLF_prev;
    double velyRF = phaseRF - phaseRF_prev;
    double velyLM = phaseLM - phaseLM_prev;
    double velyRM = phaseRM - phaseRM_prev;
    double velyLR = phaseLR - phaseLR_prev;
    double velyRR = phaseRR - phaseRR_prev;    
    
    //printf("velyLF %f\n",velyLF);
    
    double thresh = 0.0;
    
    if(velyLF > thresh) 
      wb_motor_set_position(motors[8],DOWN);
    else   
      wb_motor_set_position(motors[8],UP);  
      
    if(velyRF < -thresh) 
      wb_motor_set_position(motors[11],DOWN);
    else   
      wb_motor_set_position(motors[11],UP);  

    if(velyLM > thresh) 
      wb_motor_set_position(motors[7],DOWN);
    else   
      wb_motor_set_position(motors[7],UP);  

    if(velyRM < -thresh) 
      wb_motor_set_position(motors[10],DOWN);
    else   
      wb_motor_set_position(motors[10],UP);  

    if(velyLR > thresh) 
      wb_motor_set_position(motors[6],DOWN);
    else   
      wb_motor_set_position(motors[6],UP);  

    if(velyRR < -thresh) 
      wb_motor_set_position(motors[9],DOWN);
    else   
      wb_motor_set_position(motors[9],UP);  
                                       
  }
  
  writeOctaveTail(logFile);

  wb_robot_cleanup();

  return 0;
}
