// ============================================================================
//
// CBP_VderPol.c
// Hexapod CPG
// CBP 01-12-20
//		
// ============================================================================

#include <math.h>
#include <stdio.h>
#include "CBP_VderPol.h"

// defines for the ODEs Use labels instead of numeric indices
#define LFx 0
#define LFz 1
#define RFx 2
#define RFz 3
#define LMx 4
#define LMz 5
#define RMx 6
#define RMz 7
#define LRx 8
#define LRz 9
#define RRx 10
#define RRz 11

// Parameters
double alpha, beta, mu, omegaSW, omegaST, a, r, omega;
// Coupling matrioes angles
double theta_12, theta_13, theta_24, theta_34, theta_35, theta_46, theta_56;
double theta_21, theta_31, theta_42, theta_43, theta_53, theta_64, theta_65;
// Oscillator 'radii'
double r_LF, r_RF, r_LM, r_RM, r_LR, r_RR;
// Coupling strengths 'into'
double K_LF_x, K_LF_z, K_RF_x, K_RF_z, K_LM_x, K_LM_z;
double K_RM_x, K_RM_z, K_LR_x, K_LR_z, K_RR_x, K_RR_z;

void setParams(double alpha_, double beta_, double mu_, double omegaSW_, double a_ ){
	alpha = alpha_;
	beta = beta_;
	mu = mu_;
	omegaSW = omegaSW_;
	a = a_;
	omegaST = (1 - beta)*omegaSW/beta;	
}

void setThetas(double theta_12_, double theta_13_, double theta_24_, double theta_34_, double theta_35_, double theta_46_, double theta_56_){
	theta_12 =  theta_12_;
	theta_21 = -theta_12;
	theta_13 =  theta_13_;
	theta_31 = -theta_13;
	theta_24 =  theta_24_;
	theta_42 = -theta_24;
	theta_34 =  theta_34_;
	theta_43 = -theta_34;
	theta_35 =  theta_35_;
	theta_53 = -theta_35;
	theta_46 =  theta_46_;
	theta_64 = -theta_46;	
	theta_56 =  theta_56_;
	theta_65 = -theta_56;	
	 
}

// Naming neurons --------------------------------------------------------------
// LF	1	y(0), y(1)		LFx, LFz
// RF	2	y(2), y(3)		RFx, RFz
// LM	3	y(4), y(5)		LMx, LMz
// RM	4	y(6), y(7)		RMx, RMz
// LR	5	y(8), y(9)		LRx, LRz
// RR	6	y(10), y(11)	RRx, RRz

void vdPol3(double t, double y[], double dydt[]) {
	
// ------------ calculate r's ----------------------------------------------------
	r_LF = sqrt(y[LFx]*y[LFx] + y[LFz]*y[LFz]);
	r_RF = sqrt(y[RFx]*y[RFx] + y[RFz]*y[RFz]);
	r_LM = sqrt(y[LMx]*y[LMx] + y[LMz]*y[LMz]);
	r_RM = sqrt(y[RMx]*y[RMx] + y[RMz]*y[RMz]);
	r_LR = sqrt(y[LRx]*y[LRx] + y[LRz]*y[LRz]);
	r_RR = sqrt(y[RRx]*y[RRx] + y[RRz]*y[RRz]);

// ----------- calculate K's -----------------------------------------------------
	K_LF_x = -sin(theta_12)*(y[RFx]+y[RFz])/r_RF - sin(theta_13)*(y[LMx]+y[LMz])/r_LM;
	K_LF_z =  cos(theta_12)*(y[RFx]+y[RFz])/r_RF + cos(theta_13)*(y[LMx]+y[LMz])/r_LM;

	K_RF_x = -sin(theta_21)*(y[LFx]+y[LFz])/r_LF - sin(theta_24)*(y[RMx]+y[RMz])/r_RM;
	K_RF_z =  cos(theta_21)*(y[LFx]+y[LFz])/r_LF + cos(theta_24)*(y[RMx]+y[RMz])/r_RM;

	K_LM_x = -sin(theta_31)*(y[LFx]+y[LFz])/r_LF - sin(theta_34)*(y[RMx]+y[RMz])/r_RM -
			  sin(theta_35)*(y[LRx]+y[LRz])/r_LR;
	K_LM_z =  cos(theta_31)*(y[LFz]+y[RFx])/r_LF + cos(theta_34)*(y[RMx]+y[RMz])/r_RM +
			  cos(theta_35)*(y[LRx]+y[LRz])/r_LR;

	K_RM_x = -sin(theta_42)*(y[RFx]+y[RFz])/r_RF - sin(theta_43)*(y[LMx]+y[LMz])/r_LM -
			  sin(theta_46)*(y[RRx]+y[RRz])/r_RR;
	K_RM_z =  cos(theta_42)*(y[RFx]+y[RFz])/r_RF + cos(theta_43)*(y[LMx]+y[LMz])/r_LM +
			  cos(theta_46)*(y[RRx]+y[RRz])/r_RR;

	K_LR_x = -sin(theta_53)*(y[LMx]+y[LMz])/r_LM - sin(theta_56)*(y[RRx]+y[RRz])/r_RR;
	K_LR_z =  cos(theta_53)*(y[LMx]+y[LMz])/r_LM + cos(theta_56)*(y[RRx]+y[RRz])/r_RR;

	K_RR_x = -sin(theta_64)*(y[RMx]+y[RMz])/r_RM - sin(theta_65)*(y[LRx]+y[LRz])/r_LR;
	K_RR_z =  cos(theta_64)*(y[RMx]+y[RMz])/r_RM + cos(theta_65)*(y[LRx]+y[LRz])/r_LR;

// --------------------- dynamic system to be integrated -----------------------

	r = r_LF;
	omega = omegaST/(exp(-a*y[LFz]) + 1) + omegaSW/(exp(a*y[LFz]) + 1);
	dydt[LFx] = alpha*(mu - r*r)*y[LFx] - omega*y[LFz] + K_LF_x;
	dydt[LFz] = omega*y[LFx] + alpha*(mu - r*r)*y[LFz] + K_LF_z;

	r = r_RF;
	omega = omegaST/(exp(-a*y[RFz]) + 1) + omegaSW/(exp(a*y[RFz]) + 1);
	dydt[RFx] = alpha*(mu - r*r)*y[RFx] - omega*y[RFz] + K_RF_x;
	dydt[RFz] = omega*y[RFx] + alpha*(mu - r*r)*y[RFz] + K_RF_z;

	r = r_LM;
	omega = omegaST/(exp(-a*y[LMz]) + 1) + omegaSW/(exp(a*y[LMz]) + 1);
	dydt[LMx] = alpha*(mu - r*r)*y[LMx] - omega*y[LMz] + K_LM_x;
	dydt[LMz] = omega*y[LMx] + alpha*(mu - r*r)*y[LMz] + K_LM_z;

	r = r_RM;
	omega = omegaST/(exp(-a*y[RMz]) + 1) + omegaSW/(exp(a*y[RMz]) + 1);
	dydt[RMx] = alpha*(mu - r*r)*y[RMx] - omega*y[RMz] + K_RM_x;
	dydt[RMz] = omega*y[RMx] + alpha*(mu - r*r)*y[RMz] + K_RM_z;

	r = r_LR;
	omega = omegaST/(exp(-a*y[LRz]) + 1) + omegaSW/(exp(a*y[LRz]) + 1);
	dydt[LRx]  = alpha*(mu - r*r)*y[LRx] - omega*y[LRz] + K_LR_x;
	dydt[LRz] = omega*y[LRx] + alpha*(mu - r*r)*y[LRz] + K_LR_z;

	r = r_RR;
	omega = omegaST/(exp(-a*y[RRz]) + 1) + omegaSW/(exp(a*y[RRz]) + 1);
	dydt[RRx] = alpha*(mu - r*r)*y[RRx] - omega*y[RRz] + K_RR_x;
	dydt[RRz] = omega*y[RRx] + alpha*(mu - r*r)*y[RRz] + K_RR_z;

  return;
}

