// ============================================================================
//
// CBP_VderPol.c
//
// Quadruped rhs.
//
// ============================================================================

#include <math.h>
#include <stdio.h>
#include "CBP_VderPol.h"

double a,b,mu,omega;
double g, r0;
double K_LF[2],K_LH[2],K_RH[2],K_RF[2];
double theta1, theta2, theta3;

void setParams3(double g_, double r0_, double omega_, double t1, double t2, double t3){
  g = g_;
  r0 = r0_;
  omega = omega_;
  theta1 = t1;
  theta2 = t2;
  theta3 = t3;
}

// Oscillator labelling
// Left Front:	y[0], y[1], K1, u1
// Left Hind:	y[2], y[3], K2, u2
// Right Hind:	y[4], y[5], K3, u3
// Right Front:	y[6], y[7], K4, u4

void vdPol3(double t, double y[], double dydt[]) {
  double r;

  double u1[2];
  double u2[2];
  double u3[2];
  double u4[2]; 
    
  u1[0] = y[0]; // Left Front
  u1[1] = y[1];
  u2[0] = y[2]; // Left Hind
  u2[1] = y[3];
  u3[0] = y[4]; // Right Hind
  u3[1] = y[5];
  u4[0] = y[6]; // Right Front
  u4[1] = y[7]; 

  calcLF(theta1,theta2,theta3,u2,u3,u4,K_LF);  
  calcLH(theta1,theta2,theta3,u1,u3,u4,K_LH);
  calcRH(theta1,theta2,theta3,u1,u2,u4,K_RH);
  calcRF(theta1,theta2,theta3,u1,u2,u3,K_RF);
  

  // Left Front oscillator
  r = sqrt(y[0]*y[0] + y[1]*y[1]); 
  dydt[0] =  g*(r0/r - 1.0)*y[0] - omega*y[1] + 1.0*K_LF[0];
  dydt[1] =  g*(r0/r - 1.0)*y[1] + omega*y[0] + 1.0*K_LF[1];
  
  // Left Hind oscillator
  r = sqrt(y[2]*y[2] + y[3]*y[3]); 
  dydt[2] =  g*(r0/r - 1.0)*y[2] - omega*y[3] + 1.0*K_LH[0];
  dydt[3] =  g*(r0/r - 1.0)*y[3] + omega*y[2] + 1.0*K_LH[1];  

  // Right Hind oscillator
  r = sqrt(y[4]*y[4] + y[5]*y[5]); 
  dydt[4] =  g*(r0/r - 1.0)*y[4] - omega*y[5] + 1.0*K_RH[0];
  dydt[5] =  g*(r0/r - 1.0)*y[5] + omega*y[4] + 1.0*K_RH[1];  

  // Right Front oscillator
  r = sqrt(y[6]*y[6] + y[7]*y[7]); 
  dydt[6] =  g*(r0/r - 1.0)*y[6] - omega*y[7] + 1.0*K_RF[0];
  dydt[7] =  g*(r0/r - 1.0)*y[7] + omega*y[6] + 1.0*K_RF[1]; 
  
  //printf ( "%12f  %12f  %12f  %12f  %12f\n",  t, K1[0], K1[1], K2[0], K2[1] );

  return;
}

void rotate(double theta, double vec[], double * rot ){
	*rot = (*vec)*cos(theta) - *(vec + 1)*sin(theta);
	*(rot + 1) = (*vec)*sin(theta) + *(vec + 1)*cos(theta);
}
void add(double vec1[], double vec2[], double * sum) {
	*sum = *vec1 + *vec2;
	*(sum+1) = *(vec1+1) + *(vec2+1);
}

void calcLF(double theta1, double theta2, double theta3, double u2[], double u3[], double u4[], double K[]){
	double rot[2] = {0,0};
	*K = 0.0;
	*(K+1) = 0.0;
	rotate(-theta1,u2,rot);	
	add(K,rot,K);
	rotate(theta2,u3,rot);
	add(K,rot,K);
	rotate(-theta3,u4,rot);
	add(K,rot,K);
	//printf("u2 = %f %f, K = %f %f\n",u2[0],u2[1],K[0],K[1]);
}

void calcLH(double theta1, double theta2, double theta3, double u1[], double u3[], double u4[], double K[]){
	double rot[2];
	*K = 0.0;
	*(K+1) = 0.0;		
	rotate(theta1,u1,rot);
	add(K,rot,K);
	rotate(-theta3,u3,rot);
	add(K,rot,K);
	rotate(-theta2,u4,rot);
	add(K,rot,K);
}

void calcRH(double theta1, double theta2, double theta3, double u1[], double u2[], double u4[], double K[]){
	double rot[2];
	*K = 0.0;
	*(K+1) = 0.0;	
	rotate(-theta2,u1,rot);
	add(K,rot,K);
	rotate(theta3,u2,rot);
	add(K,rot,K);
	rotate(theta1,u4,rot);
	add(K,rot,K);
}

void calcRF(double theta1, double theta2, double theta3, double u1[], double u2[], double u3[], double K[]){
	double rot[2];
	*K = 0.0;
	*(K+1) = 0.0;	
	rotate(theta3,u1,rot);
	add(K,rot,K);
	rotate(theta2,u2,rot);
	add(K,rot,K);
	rotate(-theta1,u3,rot);
	add(K,rot,K);
}

