// ===========================================================================
//
// CBP_Quadruped_2.c
//
// CBP orig. 15-11-20
// CBP 03-03-21
// 15-03-21 RELEASE
//
// ===========================================================================

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <webots/keyboard.h>
#include <webots/motor.h>
#include <webots/robot.h>
#include <webots/gps.h>

// ODE Solvers
#include "rkf45.h"
#include "CBP_Quadruped_2_rhs.h"
#include "CBP_Helper_1.h"

// control step duration in milliseconds
#define TIME_STEP 8
#define NEQN 14

// Define gaits
#define TROT 1
#define WALK 2
#define BOUND 3
#define PACE 4
#define TRANSVERSE_GALLOP 5
#define ROTATORY_GALLOP 6
#define CANTER 7
#define PRONK 8

int main(int argc, char **argv) {
 
  WbDeviceTag motors[6];
  const char *MOTOR_NAMES[] = {"hip0", "hip1", "hip2", "hip3", "spine", NULL};
          
  // rkf stuff
  double abserr;
  int flag;
  double relerr;
  double t;
  double t_out;
  double y[NEQN];
  double dydt[NEQN];

  double theta12=0, theta23=0, theta34=0;  
  
  // for debugging
  double omega, phase, tau,k;
  FILE* logFile;
  FILE* exptFile;
    
  const char *name;  
  char fname[10];
    
  // GPS stuff
  WbDeviceTag gps;
  double x=0, xPrev, yy=0, yyPrev, speed;
  double avSpeed;
  int count=0;
  int gait;
  double pauseTime;
  
  // Get gait as command line number
  gait = atoi(argv[1]);
    
  wb_robot_init();
    
  // File stuff
  name = wb_robot_get_name();
  sprintf(fname,"Robot_%s.m",name);
  logFile = fopen(fname,"w");
  writeOctaveHead(logFile);
  sprintf(fname,"%s.txt",name);
  printf("%s\n",fname);
  exptFile = fopen(fname,"w");  
  
  pauseTime = 5.0;
  
  // Set up the phase delays for each different gait
  // Choose pauseTime from inspection of phase plots
  switch(gait) {
    case TROT:
      // Here the trot parameters
      theta12 = 0;
      theta23 = 0;
      theta34 = 0;
    break;
      
    case WALK:
      // Here the walk parameters
      theta12 = 0;
      theta23 = 0;
      theta34 = 0;      
    break;
    
    case BOUND:
      // Bound         


    break;
    
    case PACE:
      // Pace         


    break;
    
    case TRANSVERSE_GALLOP:
      // TG      


    break;
    
    case ROTATORY_GALLOP:
      // RG      


      pauseTime = 6.0;
     break;   
 
     case CANTER:
      // Canter   


     break;  
    
     case PRONK:
      // Pronk  


      pauseTime = 2.0;      
     break;        
  }     
       
  // get motor device tags
  int i;
  for (i = 0; MOTOR_NAMES[i]; i++) {
    motors[i] = wb_robot_get_device(MOTOR_NAMES[i]);
  }

  // GPS for speed calculations
  gps = wb_robot_get_device("gps");
  wb_gps_enable(gps, TIME_STEP); 
  avSpeed = 0; 
  
  // Calculate the epsilon
  abserr = sqrt ( r8_epsilon ( ) );
  relerr = sqrt ( r8_epsilon ( ) );
  printf("Root epsilon: %e\n",abserr);
  
  flag = 1;

  // Set some initial conditions
  y[0] = 1/sqrt(2);
  y[1] = 1/sqrt(2);
  
  for(int i = 2; i < NEQN;i++) 
    y[i] = 0.0;
  
  // Send over the parameters    
  setParams(10,1.0,1.7*2.0*M_PI,theta12,theta23,theta34);
  
  // for debugging 
  omega = 1.7*2.0*M_PI;
  phase = theta12;
  tau = tan(phase/4.0)/omega;
  k = sqrt(tau*tau*omega*omega + 1);
  printf("tau12 = %f, k12 = %f\n",tau,k);
  phase = theta23;
  tau = tan(phase/4.0)/omega;
  k = sqrt(tau*tau*omega*omega + 1);
  printf("tau23 = %f, k23 = %f\n",tau,k);
  phase = theta34;
  tau = tan(phase/4.0)/omega;
  k = sqrt(tau*tau*omega*omega + 1);
  printf("tau34 = %f, k34 = %f\n",tau,k);    
           
  t = 0.0;
  t_out = TIME_STEP/1000.0; 
    
  while (wb_robot_step(TIME_STEP) != -1) {
                      
    flag = r8_rkf45 (rhs_2, NEQN, y, dydt, &t, t_out, &relerr, abserr, flag );
      
      
  //  printf ( "_2: %4d  %12f  %12f  %12f  %12f  %12f\n", flag, t, y[0], y[5], y[9], y[13] );
    if(flag == 4) {
      printf("WARNING:%s: flag 4 at t = %f\n",name,t);
      flag = 2;
    }
    if(flag == 7) {
      printf("WARNING:%s: flag 7 at t = %f\n",name,t);
      flag = 2; 
    } 
     
    t_out = t + TIME_STEP/1000.0;
    
    double phaseLF = y[0];
    double phaseLH = y[5];
    double phaseRH = y[9];
    double phaseRF = y[13];
    
    writeData4ToOctave(logFile,t,phaseLF,phaseLH,phaseRH,phaseRF);    
      
      // Chosen empirically     
    double scalingF = 0.25;
    double scalingH = 0.15;
    double offsetH = -0.2;
    double offsetF = 0.0;
    
    // latest values - these are better and
    // influcence how gait develops due to
    // interaction with physical world.
    // OK for trot but not for walk
    /*
     scalingF = 0.35;
     scalingH = 0.25;
     offsetH  = -0.1;
     offsetF  = 0.0;    
    */
      
    phaseLF = phaseLF*scalingF + offsetF;
    phaseLH = phaseLH*scalingH + offsetH;      
    phaseRH = phaseRH*scalingH + offsetH;      
    phaseRF = phaseRF*scalingF + offsetF;      

    if(t > pauseTime) {
      wb_motor_set_position(motors[2], phaseLF);
      wb_motor_set_position(motors[3], phaseLH);      
      wb_motor_set_position(motors[1], phaseRH);    
      wb_motor_set_position(motors[0], phaseRF);
    }
       
   // GPS for speed calc
    xPrev = x;
    yyPrev = yy;
    const double *gpsVals = wb_gps_get_values(gps);  
    x = gpsVals[0];
    yy = gpsVals[2];    
    speed = sqrt(pow(x - xPrev,2) + pow(yy - yyPrev,2))/(TIME_STEP/1000.0);
   // printf("%s speed = %f\n",name,speed); 
    
    if(t > 5) {
      avSpeed += speed;
      count++;
      if(count > 100) {
        avSpeed = avSpeed/(double)count;
        printf("%s Average speed = %f\n",name,avSpeed);
        fprintf(exptFile," t = %3f, avSpeed = %f\n",t,avSpeed);
        count = 0;
        avSpeed = 0;
      }
    }  
	     

  }
  
  writeOctaveTail(logFile);
  fflush(logFile);
  fclose(logFile);
  fflush(exptFile);
  fclose(exptFile);
  
  wb_robot_cleanup();

  return 0;
}
